
from typing import Any, TypeVar, Type, Dict, Mapping, MutableSet, FrozenSet, \
    NamedTuple, Optional, Iterable, List, Tuple, SupportsBytes
from weakref import WeakSet

from .concurrent import Event
from .flags import FlagOp, PermanentFlags, SessionFlags
from .parsing.command import Command
from .parsing.primitives import ListP, Number
from .parsing.response import Response, ResponseBye
from .parsing.response.code import UidValidity
from .parsing.response.specials import ExistsResponse, RecentResponse, \
    ExpungeResponse, FetchResponse
from .parsing.specials import FetchAttribute, Flag, SequenceSet

__all__ = ['SelectedSet', 'SelectedSnapshot', 'SelectedMailbox', 'SelectedT']

#: Type variable with an upper bound of :class:`SelectedMailbox`.
SelectedT = TypeVar('SelectedT', bound='SelectedMailbox')

_Flags = FrozenSet[Flag]
_Message = Tuple[int, _Flags]
_flags_attr = FetchAttribute(b'FLAGS')
_uid_attr = FetchAttribute(b'UID')


class SelectedSet:
    """Maintains a weak set of :class:`SelectedMailbox` objects that exist for
    a mailbox, across all sessions. This is useful for assigning the
    ``\\Recent`` flag, as well as notifying other sessions about updates.

    Args:
        updated: The event to notify when updates occur. Defaults to a new
            event using :mod:`asyncio` concurrency primitives.

    """

    __slots__ = ['set', '_updated']

    def __init__(self, updated: Event = None) -> None:
        super().__init__()
        self.set: MutableSet['SelectedMailbox'] = WeakSet()
        self._updated = updated or Event.for_asyncio()

    @property
    def updated(self) -> Event:
        """The event to notify when updates occur."""
        return self._updated

    @property
    def any_selected(self) -> Optional['SelectedMailbox']:
        """A single, random object in the set of selected mailbox objects.
        Selected mailbox object's marked :attr:`~SelectedMailbox.readonly`
        will not be chosen.

        """
        for selected in self.set:
            if not selected.readonly:
                return selected
        return None


class SelectedSnapshot(NamedTuple):
    """Holds a snapshot of the selected mailbox as of the last fork.

    Args:
        uid_validity: The UID validity of the selected mailbox.
        next_uid: The predicted next message UID of the mailbox.
        recent: The number of messages in the mailbox with ``\\Recent``.
        messages: The message UIDs mapped to flags.

    """

    uid_validity: Optional[int]
    next_uid: int
    recent: int
    messages: Dict[int, _Flags]

    @property
    def exists(self) -> int:
        """The total number of messages in the mailbox."""
        return len(self.messages)


class SelectedMailbox:
    """Manages the updates to the current selected mailbox from other
    concurrent sessions.

    The current state of the selected mailbox will be written to this object by
    the backend implementation during each operation.  Then, when the operation
    completes, call :meth:`.fork` to make a fresh copy of the object and any
    untagged responses that should be added to the response.

    Args:
        name: The name of the selected mailbox.
        readonly: Indicates the mailbox is selected as read-only.
        permanent_flags: The defined permanent flags for the mailbox.
        session_flags: Session-only flags for the mailbox.
        selected_set: The ``self`` object and subsequent forked copies will be
            kept in in this set.

    """

    def __init__(self, name: str, readonly: bool,
                 permanent_flags: PermanentFlags,
                 session_flags: SessionFlags,
                 selected_set: SelectedSet = None,
                 **kwargs: Any) -> None:
        super().__init__()
        self._name = name
        self._readonly = readonly
        self._permanent_flags = permanent_flags
        self._session_flags = session_flags
        self._selected_set = selected_set
        self._kwargs = kwargs
        self._uid_validity: Optional[int] = None
        self._next_uid = 1
        self._is_deleted = False
        self._messages: Dict[int, _Flags] = {}
        self._hashed: Optional[Mapping[int, int]] = None
        self._hide_expunged = False
        self._snapshot: Optional[SelectedSnapshot] = None
        if selected_set:
            selected_set.set.add(self)

    @property
    def name(self) -> str:
        """The name of the selected mailbox."""
        return self._name

    @property
    def uid_validity(self) -> Optional[int]:
        """The UID validity of the selected mailbox."""
        return self._uid_validity

    @uid_validity.setter
    def uid_validity(self, uid_validity: int) -> None:
        self._uid_validity = uid_validity

    @property
    def next_uid(self) -> int:
        """The predicted next message UID value of the mailbox."""
        return self._next_uid

    @next_uid.setter
    def next_uid(self, next_uid: int) -> None:
        self._next_uid = next_uid

    @property
    def readonly(self) -> bool:
        """Indicates the mailbox is selected as read-only."""
        return self._readonly

    @property
    def exists(self) -> int:
        """The total number of messages in the mailbox."""
        return len(self._messages)

    @property
    def recent(self) -> int:
        """The number of messages in the mailbox with ``\\Recent``."""
        recent_uids = self.session_flags.recent_uids
        if self._hide_expunged:
            return len(recent_uids)
        else:
            current_uids = frozenset(self._messages.keys())
            return len(recent_uids & current_uids)

    @property
    def permanent_flags(self) -> PermanentFlags:
        """The defined permanent flags for the mailbox."""
        return self._permanent_flags

    @property
    def session_flags(self) -> SessionFlags:
        """Session-only flags for the mailbox."""
        return self._session_flags

    @property
    def snapshot(self) -> SelectedSnapshot:
        """A snapshot of the selected mailbox as of the last fork."""
        if not self._snapshot:
            raise RuntimeError()  # Must call fork() first.
        return self._snapshot

    @property
    def kwargs(self) -> Mapping[str, Any]:
        """Add keywords arguments to copy construction during :meth:`.fork`."""
        return {}

    def set_deleted(self) -> None:
        """Marks the selected mailbox as having been deleted."""
        self._is_deleted = True

    def add_messages(self, *messages: _Message) -> None:
        """Add a message that exists in the mailbox. Shortcut for
        :meth:`.add_message`.

        Args:
            messages: The messages to add, each a tuple of UID and permanent
                flags.

        """
        for uid, permanent_flags in messages:
            all_flags = permanent_flags | self.session_flags[uid]
            self._messages[uid] = all_flags

    def silence(self, uids: Iterable[int], flag_set: FrozenSet[Flag],
                flag_op: FlagOp) -> None:
        """Runs the flags update against the flags at the last
        :meth:`.fork`, to prevent untagged FETCH responses unless other
        updates have occurred.

        For example, if a session adds ``\\Deleted`` and calls this method,
        the FETCH response will be silenced. But if another added ``\\Seen``
        at the same time, the FETCH response will be sent.

        Args:
            uids: The set of message UIDs.
            flag_set: The set of flags for the update operation.
            flag_op: The mode to change the flags.

        """
        permanent_flags = self.permanent_flags & flag_set
        session_flags = self.session_flags & flag_set
        defined_flags = permanent_flags | session_flags
        messages = self.snapshot.messages
        for uid in uids:
            orig_flags = messages.get(uid)
            if orig_flags is not None:
                new_flags = flag_op.apply(orig_flags, defined_flags)
                self.snapshot.messages[uid] = new_flags

    def hide_expunged(self) -> None:
        """No untagged ``EXPUNGE`` responses will be generated, and message
        sequence numbers will not be adjusted, until the next :meth:`.fork`.

        """
        self._hide_expunged = True

    def iter_set(self, seq_set: SequenceSet, uids: FrozenSet[int]) \
            -> Iterable[Tuple[int, int]]:
        """Iterate through the given sequence set based on the message state at
        the last fork.

        Args:
            seq_set: Sequence set to convert to UID set.
            uids: The current set of UIDs in the mailbox.

        """
        if self._hide_expunged:
            snapshot_uids = frozenset(self.snapshot.messages.keys())
            all_uids = uids | snapshot_uids
        else:
            all_uids = uids
        sorted_uids = sorted(all_uids)
        if seq_set.uid:
            try:
                max_uid = sorted_uids[-1]
            except IndexError:
                max_uid = 0
            all_idx = frozenset(seq_set.iter(max_uid))
        else:
            all_idx = frozenset(seq_set.iter(len(all_uids)))
        for seq, uid in enumerate(sorted_uids, 1):
            idx = uid if seq_set.uid else seq
            if idx in all_idx:
                yield (seq, uid)

    def fork(self: SelectedT, command: Command) \
            -> Tuple[SelectedT, Iterable[Response]]:
        """Compares the state of the current object to that of the last fork,
        returning the untagged responses that reflect any changes. A new copy
        of the object is also returned, ready for the next command.

        Args:
            command: The command that was finished.

        """
        cls: Type[SelectedT] = type(self)
        copy = cls(self.name, self.readonly, self._permanent_flags,
                   self._session_flags, self._selected_set, **self.kwargs)
        if self._hide_expunged and self._snapshot:
            messages = self.snapshot.messages.copy()
            messages.update(self._messages)
        else:
            messages = self._messages
        copy._snapshot = SelectedSnapshot(self.uid_validity, self.next_uid,
                                          self.recent, messages)
        if self._selected_set:
            self._selected_set.set.discard(self)
            self._selected_set.set.add(copy)
        if self._snapshot:
            return copy, self._compare(command)
        else:
            return copy, []

    def _compare(self, command: Command) -> Iterable[Response]:
        is_uid: bool = getattr(command, 'uid', False)
        before_uidval, _, before_recent, before = self.snapshot
        current = self._messages
        uidval = self.uid_validity
        if self._is_deleted:
            yield ResponseBye(b'Selected mailbox deleted.')
            return
        elif uidval is not None and before_uidval != uidval:
            yield ResponseBye(b'UID validity changed.', UidValidity(uidval))
            return
        before_uids = frozenset(before.keys())
        current_uids = frozenset(current.keys())
        expunged_uids = before_uids - current_uids
        both_uids = before_uids & current_uids
        new_uids = current_uids - before_uids
        both: List[Tuple[int, int]] = []
        new: List[Tuple[int, int]] = []
        if self._hide_expunged:
            for uid in expunged_uids:
                current[uid] = before[uid]
            current_uids = frozenset(current.keys())
        else:
            expunged: List[int] = []
            sorted_before_uids = sorted(before_uids)
            for seq, uid in enumerate(sorted_before_uids, 1):
                if uid in expunged_uids:
                    del self.session_flags[uid]
                    expunged.append(seq)
            for seq in reversed(expunged):
                yield ExpungeResponse(seq)
        for seq, uid in enumerate(sorted(current_uids), 1):
            if uid in both_uids:
                both.append((seq, uid))
            elif uid in new_uids:
                new.append((seq, uid))
        if new_uids:
            yield ExistsResponse(len(current_uids))
        if before_recent != self.recent:
            yield RecentResponse(self.recent)
        for seq, uid in both:
            before_flags = before[uid]
            current_flags = current[uid]
            if hash(before_flags) != hash(current_flags):
                fetch_data: Dict[FetchAttribute, SupportsBytes] = {
                    _flags_attr: ListP(current_flags, sort=True)}
                if is_uid:
                    fetch_data[_uid_attr] = Number(uid)
                yield FetchResponse(seq, fetch_data)
        for seq, uid in new:
            current_flags = current[uid]
            fetch_data = {_flags_attr: ListP(current_flags, sort=True)}
            if is_uid:
                fetch_data[_uid_attr] = Number(uid)
            yield FetchResponse(seq, fetch_data)



from typing import Tuple

from .. import Params, Parseable, Space
from ..exceptions import NotParseable, InvalidContent
from ..primitives import Atom

__all__ = ['StatusAttribute']


class StatusAttribute(Parseable[bytes]):
    """Represents a status attribute from an IMAP stream.

    Args:
        status: The status attribute string.

    """

    #: The set of valid status attributes.
    valid_statuses = {b'MESSAGES', b'RECENT', b'UIDNEXT', b'UIDVALIDITY',
                      b'UNSEEN'}

    def __init__(self, status: bytes) -> None:
        super().__init__()
        status = status.upper()
        if status not in self.valid_statuses:
            raise ValueError(status)
        self.status = status

    @property
    def value(self) -> bytes:
        """The status attribute string."""
        return self.status

    @classmethod
    def parse(cls, buf: bytes, params: Params) \
            -> Tuple['StatusAttribute', bytes]:
        try:
            _, buf = Space.parse(buf, params)
        except NotParseable:
            pass
        atom, after = Atom.parse(buf, params)
        try:
            return cls(atom.value), after
        except ValueError:
            raise InvalidContent(buf)

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other) -> bool:
        if isinstance(other, StatusAttribute):
            return self.value == other.value
        return super().__eq__(other)

    def __bytes__(self) -> bytes:
        return self.value




from typing import TypeVar, Type, Any, Optional, List, Dict, Tuple

from ...bytes import MaybeBytes, BytesFormat

__all__ = ['ResponseCode', 'Response', 'ResponseContinuation', 'ResponseBad',
           'ResponseNo', 'ResponseOk', 'ResponseBye', 'ResponsePreAuth',
           'ResponseT']

#: Type variable with an upper bound of :class:`Response`.
ResponseT = TypeVar('ResponseT', bound='Response')

_Mergeable = Dict[Tuple[Type['Response'], Any], int]


class ResponseCode:
    """Base class for response codes that may be returned along with IMAP
    server responses.

    """

    def __bytes__(self) -> bytes:
        raise NotImplementedError

    @classmethod
    def of(cls, code: Optional[MaybeBytes]) -> Optional['ResponseCode']:
        """Build and return an anonymous response code object.

        Args:
            code: The code string, without square brackets.

        """
        if code is not None:
            return _AnonymousResponseCode(code)
        else:
            return None


class _AnonymousResponseCode(ResponseCode):

    def __init__(self, code: MaybeBytes) -> None:
        super().__init__()
        self.code = code

    def __bytes__(self) -> bytes:
        return BytesFormat(b'[%b]') % self.code


class Response:
    """Base class for all responses sent from the server to the client. These
    responses may be sent unsolicited (e.g. idle timeouts) or in response to a
    tagged command from the client.

    Args:
        tag: The tag bytestring of the associated command, a plus (``+``) to
            indicate a continuation requirement, or an asterisk (``*``) to
            indicate an untagged response.
        text: The response text.
        code: Optional response code.

    Attributes:
        tag: The tag bytestring.

    """

    #: The condition bytestring, e.g. ``OK``.
    condition: Optional[bytes] = None

    def __init__(self, tag: MaybeBytes, text: MaybeBytes = None,
                 code: ResponseCode = None) -> None:
        super().__init__()
        self.tag = bytes(tag)
        self.code = code
        self._text = text or b''
        self._untagged: List['Response'] = []
        self._mergeable: _Mergeable = {}
        self._raw: Optional[bytes] = None

    @property
    def text(self) -> bytes:
        """The response text."""
        if self.condition:
            if self.code:
                return BytesFormat(b'%b %b %b') \
                    % (self.condition, self.code, self._text)
            else:
                return BytesFormat(b'%b %b') % (self.condition, self._text)
        else:
            return bytes(self._text)

    def add_untagged(self, *responses: 'Response') -> None:
        """Add an untagged response. These responses are shown before the
        parent response.

        Args:
            responses: The untagged responses to add.

        """
        for resp in responses:
            try:
                merge_key = resp.merge_key
            except TypeError:
                self._untagged.append(resp)
            else:
                key = (type(resp), merge_key)
                try:
                    untagged_idx = self._mergeable[key]
                except KeyError:
                    untagged_idx = len(self._untagged)
                    self._mergeable[key] = untagged_idx
                    self._untagged.append(resp)
                else:
                    merged = self._untagged[untagged_idx].merge(resp)
                    self._untagged[untagged_idx] = merged
        self._raw = None

    def add_untagged_ok(self, text: MaybeBytes,
                        code: Optional[ResponseCode] = None) -> None:
        """Add an untagged ``OK`` response.

        See Also:
            :meth:`.add_untagged`, :class:`ResponseOk`

        Args:
            text: The response text.
            code: Optional response code.

        """
        response = ResponseOk(b'*', text, code)
        self.add_untagged(response)

    @property
    def is_terminal(self) -> bool:
        """True if the response contained an untagged ``BYE`` response
        indicating that the session should be terminated.

        """
        for resp in self._untagged:
            if resp.is_terminal:
                return True
        return False

    @property
    def merge_key(self) -> Any:
        """Returns a hashable value which can be compared to other
        :attr:`.merge_key` values of the same response type to see if the
        two responses can be merged.

        Raises:
            TypeError: This response type may not be merged.

        """
        raise TypeError(self)

    def merge(self: ResponseT, other: ResponseT) -> ResponseT:
        """Return a copy of this response with the other response merged in.

        Args:
            other: The other response to merge.

        Raises:
            TypeError: This response type may not be merged.
            ValueError: The two responses are not mergeable.

        """
        raise TypeError(self)

    def __bytes__(self) -> bytes:
        if self._raw is not None:
            return self._raw
        resp_line = BytesFormat(b'%b %b\r\n') % (self.tag, self.text)
        self._raw = BytesFormat(b'').join(self._untagged, [resp_line])
        return self._raw


class ResponseContinuation(Response):
    """Class used for server responses that indicate a continuation
    requirement. This is when the server needs more data from the client to
    finish handling the command. The ``AUTHENTICATE`` command and any command
    that uses a literal string argument will send this response as needed.

    Args:
        text: The continuation text.

    """

    def __init__(self, text: MaybeBytes) -> None:
        super().__init__(b'+', text)


class ResponseBad(Response):
    """``BAD`` response indicating the server encountered a protocol-related
    error in responding to the command.

    Args:
        tag: The tag bytestring to associate the response to a command.
        text: The response text.
        code: Optional response code.

    """

    condition = b'BAD'

    def __init__(self, tag: MaybeBytes, text: MaybeBytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__(tag, text, code)


class ResponseNo(Response):
    """``NO`` response indicating the server successfully parsed the command
    but failed to execute it successfully.

    Args:
        tag: The tag bytestring to associate the response to a command.
        text: The response text.
        code: Optional response code.

    """

    condition = b'NO'

    def __init__(self, tag: MaybeBytes, text: MaybeBytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__(tag, text, code)


class ResponseOk(Response):
    """``OK`` response indicating the server successfully parsed and executed
    the command.

    Args:
        tag: The tag bytestring to associate the response to a command.
        text: The response text.
        code: Optional response code.

    """

    condition = b'OK'

    def __init__(self, tag: MaybeBytes, text: MaybeBytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__(tag, text, code)


class ResponseBye(Response):
    """``BYE`` response indicating that the server will be closing the
    connection immediately after sending the response is sent. This may be sent
    in response to a command (e.g. ``LOGOUT``) or unsolicited.

    Args:
        text: The reason for disconnection.
        code: Optional response code.

    """

    condition = b'BYE'

    def __init__(self, text: MaybeBytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__(b'*', text, code)

    @property
    def is_terminal(self) -> bool:
        """This response is always terminal."""
        return True


class ResponsePreAuth(Response):
    """``PREAUTH`` response during server greeting to indicate the client is
    already logged in.

    Args:
        tag: The tag bytestring to associate the response to a command.
        text: The response text.
        code: Optional response code.

    """

    condition = b'PREAUTH'

    def __init__(self, tag: MaybeBytes, text: MaybeBytes,
                 code: Optional[ResponseCode] = None) -> None:
        super().__init__(tag, text, code)




import unittest

from pymap.flags import FlagOp, PermanentFlags, SessionFlags
from pymap.parsing.command.select import SearchCommand, UidSearchCommand
from pymap.parsing.response import ResponseOk
from pymap.parsing.specials.flag import Seen, Flagged
from pymap.selected import SelectedMailbox


class TestSelectedMailbox(unittest.TestCase):

    def setUp(self) -> None:
        self.response = ResponseOk(b'.', b'testing')

    @classmethod
    def new_selected(cls) -> SelectedMailbox:
        return SelectedMailbox('test', False,
                               PermanentFlags(Seen, Flagged),
                               SessionFlags())

    @property
    def command(self) -> SearchCommand:
        return SearchCommand(b'.', [], None)

    @property
    def uid_command(self) -> SearchCommand:
        return UidSearchCommand(b'.', [], None)

    def test_add_untagged_recent_equal(self) -> None:
        selected = self.new_selected()
        selected.session_flags.add_recent(1)
        selected.session_flags.add_recent(2)
        selected.add_messages((1, frozenset()),
                              (2, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((1, frozenset()),
                            (2, frozenset()))
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_recent_increase(self) -> None:
        selected = self.new_selected()
        selected.session_flags.add_recent(1)
        selected.session_flags.add_recent(2)
        selected.add_messages((1, frozenset()),
                              (2, frozenset()),
                              (3, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.session_flags.add_recent(3)
        forked.add_messages((1, frozenset()),
                            (2, frozenset()),
                            (3, frozenset()))
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 RECENT\r\n'
                         b'* 3 FETCH (FLAGS (\\Recent))\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_recent_expunge(self) -> None:
        selected = self.new_selected()
        selected.session_flags.add_recent(1)
        selected.session_flags.add_recent(2)
        selected.add_messages((1, frozenset()),
                              (2, frozenset()),
                              (3, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((1, frozenset()))
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 EXPUNGE\r\n'
                         b'* 2 EXPUNGE\r\n'
                         b'* 1 RECENT\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_equal(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()),
                              (2, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((1, frozenset()),
                            (2, frozenset()))
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_fetch(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()),
                              (2, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((1, frozenset()),
                            (2, frozenset([Seen])),
                            (3, frozenset([Seen, Flagged])),
                            (4, frozenset()))
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 4 EXISTS\r\n'
                         b'* 2 FETCH (FLAGS (\\Seen))\r\n'
                         b'* 3 FETCH (FLAGS (\\Flagged \\Seen))\r\n'
                         b'* 4 FETCH (FLAGS ())\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_fetch_uid(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()),
                              (2, frozenset()))
        forked, _ = selected.fork(self.uid_command)
        forked.add_messages((1, frozenset()),
                            (2, frozenset([Seen])),
                            (3, frozenset([Seen, Flagged])),
                            (4, frozenset()))
        _, untagged = forked.fork(self.uid_command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 4 EXISTS\r\n'
                         b'* 2 FETCH (FLAGS (\\Seen) UID 2)\r\n'
                         b'* 3 FETCH (FLAGS (\\Flagged \\Seen) UID 3)\r\n'
                         b'* 4 FETCH (FLAGS () UID 4)\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_fetch_silenced(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()),
                              (2, frozenset()),
                              (3, frozenset([Seen, Flagged])),
                              (4, frozenset([Seen, Flagged])),
                              (5, frozenset([Flagged])),
                              (6, frozenset([Seen])))
        forked, _ = selected.fork(self.uid_command)
        forked.add_messages((1, frozenset([Seen, Flagged])),
                            (2, frozenset([Seen])),
                            (3, frozenset([])),
                            (4, frozenset([Flagged])),
                            (5, frozenset([Seen, Flagged])),
                            (6, frozenset([Seen])))
        forked.silence([1, 2], frozenset([Seen]), FlagOp.ADD)
        forked.silence([3, 4], frozenset([Seen]), FlagOp.DELETE)
        forked.silence([5, 6], frozenset([Seen]), FlagOp.REPLACE)
        _, untagged = forked.fork(self.uid_command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 1 FETCH (FLAGS (\\Flagged \\Seen) UID 1)\r\n'
                         b'* 3 FETCH (FLAGS () UID 3)\r\n'
                         b'* 5 FETCH (FLAGS (\\Flagged \\Seen) UID 5)\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_expunge_hidden(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()),
                              (2, frozenset()),
                              (3, frozenset()),
                              (4, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((1, frozenset()),
                            (4, frozenset()),
                            (5, frozenset()))
        forked.hide_expunged()
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 5 EXISTS\r\n'
                         b'* 5 FETCH (FLAGS ())\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_expunge(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()),
                              (2, frozenset()),
                              (3, frozenset()),
                              (4, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((1, frozenset()),
                            (4, frozenset()))
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 EXPUNGE\r\n'
                         b'* 2 EXPUNGE\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_all(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset([Flagged])),
                              (2, frozenset()),
                              (3, frozenset()))
        forked, _ = selected.fork(self.uid_command)
        selected.session_flags.add_recent(6)
        forked.add_messages((1, frozenset([Seen, Flagged])),
                            (4, frozenset([Seen])),
                            (5, frozenset([Seen])),
                            (6, frozenset([Flagged])),
                            (7, frozenset([Seen])))
        _, untagged = forked.fork(self.uid_command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* 3 EXPUNGE\r\n'
                         b'* 2 EXPUNGE\r\n'
                         b'* 5 EXISTS\r\n'
                         b'* 1 RECENT\r\n'
                         b'* 1 FETCH (FLAGS (\\Flagged \\Seen) UID 1)\r\n'
                         b'* 2 FETCH (FLAGS (\\Seen) UID 4)\r\n'
                         b'* 3 FETCH (FLAGS (\\Seen) UID 5)\r\n'
                         b'* 4 FETCH (FLAGS (\\Flagged \\Recent) UID 6)\r\n'
                         b'* 5 FETCH (FLAGS (\\Seen) UID 7)\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_deleted_bye(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((2, frozenset()))
        forked.set_deleted()
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* BYE Selected mailbox deleted.\r\n'
                         b'. OK testing\r\n', bytes(self.response))

    def test_add_untagged_uid_validity_bye(self) -> None:
        selected = self.new_selected()
        selected.add_messages((1, frozenset()))
        forked, _ = selected.fork(self.command)
        forked.add_messages((2, frozenset()))
        forked.uid_validity = 456
        _, untagged = forked.fork(self.command)
        self.response.add_untagged(*untagged)
        self.assertEqual(b'* BYE [UIDVALIDITY 456] UID validity changed.\r\n'
                         b'. OK testing\r\n', bytes(self.response))




from . import CommandAny, CommandNoArgs

__all__ = ['CapabilityCommand', 'LogoutCommand', 'NoOpCommand']


class CapabilityCommand(CommandNoArgs, CommandAny):
    """The ``CAPABILITY`` command lists current server capabilities."""

    command = b'CAPABILITY'


class LogoutCommand(CommandNoArgs, CommandAny):
    """The ``LOGOUT`` command ends the IMAP session."""

    command = b'LOGOUT'


class NoOpCommand(CommandNoArgs, CommandAny):
    """The ``NOOP`` command does nothing, but can be used to check for state
    changes on a selected mailbox.

    """

    command = b'NOOP'




from typing import TypeVar, Tuple, AsyncIterable, AsyncIterator

__all__ = ['asyncenumerate', 'AsyncEnumerate', 'AsyncEnumerateT']

#: Type variable bound to the enumerated element in :class:`AsyncEnumerate`.
AsyncEnumerateT = TypeVar('AsyncEnumerateT')


class _AsyncEnumerateIterator(AsyncIterator[Tuple[int, AsyncEnumerateT]]):

    __slots__ = ['_iter', '_idx']

    def __init__(self, iterator: AsyncIterator[AsyncEnumerateT],
                 idx: int) -> None:
        super().__init__()
        self._iter = iterator
        self._idx = idx

    def __aiter__(self) -> AsyncIterator[Tuple[int, AsyncEnumerateT]]:
        return self

    async def __anext__(self) -> Tuple[int, AsyncEnumerateT]:
        item = await self._iter.__anext__()
        self._idx += 1
        return self._idx, item


class AsyncEnumerate(AsyncIterable[Tuple[int, AsyncEnumerateT]]):
    """Imitates Python's :func:`enumerate` with async iterators.

    Args:
        iterable: The iterable to enumerate over.
        start: The starting value of the index.

    """

    __slots__ = ['_sub_iter', '_start']

    def __init__(self, iterable: AsyncIterable[AsyncEnumerateT],
                 start: int = 0) -> None:
        super().__init__()
        self._sub_iter = iterable
        self._start = start - 1

    def __aiter__(self) -> AsyncIterator[Tuple[int, AsyncEnumerateT]]:
        sub_iter = self._sub_iter.__aiter__()
        return _AsyncEnumerateIterator(sub_iter, self._start)


# Expose AsyncEnumerator class with a public alias.
asyncenumerate = AsyncEnumerate




from abc import abstractmethod
from datetime import datetime
from typing import Optional, Iterable, Set, FrozenSet, Sequence, TypeVar
from typing_extensions import Protocol

from ..flags import FlagOp, SessionFlags
from ..parsing.response.fetch import EnvelopeStructure, BodyStructure
from ..parsing.specials import Flag

__all__ = ['Header', 'MessageInterface', 'MessageT']

#: Type variable with an upper bound of :class:`MessageInterface`.
MessageT = TypeVar('MessageT', bound='MessageInterface')


class Header(Protocol):
    """A message header value, which is convertible to a string with
    :class:`str`.

    """

    @abstractmethod
    def __str__(self) -> str:
        ...


class MessageInterface(Protocol):
    """Message metadata such as UID, permanent flags, and when the message
    was added to the system.

    """

    @property
    @abstractmethod
    def uid(self) -> int:
        """The message's unique identifier in the mailbox."""
        ...

    @property
    @abstractmethod
    def expunged(self) -> bool:
        """True if this message has been expunged from the mailbox."""
        ...

    @property
    @abstractmethod
    def permanent_flags(self) -> Set[Flag]:
        """The message's set of permanent flags."""
        ...

    @property
    @abstractmethod
    def internal_date(self) -> Optional[datetime]:
        """The message's internal date."""
        ...

    @abstractmethod
    def copy(self: MessageT, new_uid: int) -> MessageT:
        """Return a copy of the message with a new UID.

        Args:
            new_uid: The copied message UID.

        """
        ...

    @abstractmethod
    def get_flags(self, session_flags: SessionFlags = None) -> FrozenSet[Flag]:
        """Get the full set of permanent and session flags for the message.

        Args:
            session_flags: The current session flags.

        """
        ...

    @abstractmethod
    def update_flags(self, flag_set: Iterable[Flag],
                     flag_op: FlagOp = FlagOp.REPLACE) -> FrozenSet[Flag]:
        """Update the permanent flags for the message. Returns the resulting
        set of permanent flags.

        Args:
            flag_set: The set of flags for the update operation.
            flag_op: The mode to change the flags.

        """
        ...

    @abstractmethod
    def get_header(self, name: str) -> Sequence[Header]:
        """Get the values of a header from the message.

        Args:
            name: The name of the header.

        """
        ...

    @abstractmethod
    def get_headers(self, section: Iterable[int] = None,
                    subset: Iterable[str] = None,
                    inverse: bool = False) -> Optional[bytes]:
        """Get the headers from the message.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            subset: Subset of headers to get.
            inverse: If ``subset`` is given, this flag will invert it so that
                the headers *not* in ``subset`` are returned.

        """
        ...

    @abstractmethod
    def get_body(self, section: Iterable[int] = None,
                 binary: bool = False) -> Optional[bytes]:
        """Get the full body of the message part, including headers.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            binary: True if the result uses 8-bit encoding.

        """
        ...

    @abstractmethod
    def get_text(self, section: Iterable[int] = None,
                 binary: bool = False) -> Optional[bytes]:
        """Get the text of the message part, not including headers.

        The ``section`` argument can index a nested sub-part of the message.
        For example, ``[2, 3]`` would get the 2nd sub-part of the message and
        then index it for its 3rd sub-part.

        Args:
            section: Optional nested list of sub-part indexes.
            binary: True if the result uses 8-bit encoding.

        """
        ...

    @abstractmethod
    def get_size(self, section: Iterable[int] = None,
                 binary: bool = False) -> int:
        """Return the size of the message, in octets.

        Args:
            section: Optional nested list of sub-part indexes.
            binary: True if the result uses 8-bit encoding.

        """
        ...

    @abstractmethod
    def get_envelope_structure(self) -> EnvelopeStructure:
        """Build and return the envelope structure.

        See Also:
            `RFC 3501 2.3.5.
            <https://tools.ietf.org/html/rfc3501#section-2.3.5>`_

        """
        ...

    @abstractmethod
    def get_body_structure(self) -> BodyStructure:
        """Build and return the body structure.

        See Also:
            `RFC 3501 2.3.6
            <https://tools.ietf.org/html/rfc3501#section-2.3.6>`_

        """
        ...



# Copyright (c) 2018 Ian C. Good
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from setuptools import setup, find_packages  # type: ignore


with open('README.md', 'r') as fh:
    long_description = fh.read()


setup(name='pymap',
      version='0.5.1-r1',
      author='Ian Good',
      author_email='icgood@gmail.com',
      description='Lightweight, asynchronous IMAP serving in Python.',
      long_description=long_description,
      long_description_content_type='text/markdown',
      license='MIT',
      url='https://github.com/icgood/pymap/',
      packages=find_packages(),
      python_requires='~=3.7',
      install_requires=['pysasl', 'typing-extensions'],
      tests_require=['pytest', 'pytest-asyncio'],
      entry_points={'console_scripts': ['pymap = pymap.main:main'],
                    'pymap.backend': ['dict = pymap.backend.dict',
                                      'maildir = pymap.backend.maildir']},
      package_data={'pymap.backend.dict': ['demo/']},
      classifiers=['Development Status :: 3 - Alpha',
                   'Topic :: Communications :: Email :: Post-Office',
                   'Topic :: Communications :: Email :: Post-Office :: IMAP',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'License :: OSI Approved :: MIT License',
                   'Programming Language :: Python',
                   'Programming Language :: Python :: 3.7'])



"""Base implementations of the :mod:`pymap.interfaces.message` interfaces."""

from datetime import datetime
from email.generator import BytesGenerator
from email.message import EmailMessage
from email.parser import BytesParser
from email.policy import SMTP
from io import BytesIO
from typing import Any, Tuple, Optional, Iterable, Set, Dict, FrozenSet, \
    Sequence, List, NamedTuple, BinaryIO, TypeVar, Type

from .flags import FlagOp, SessionFlags
from .interfaces.message import Header, MessageInterface
from .parsing.response.fetch import EnvelopeStructure, BodyStructure, \
    MultipartBodyStructure, ContentBodyStructure, TextBodyStructure, \
    MessageBodyStructure
from .parsing.specials import ExtensionOptions
from .parsing.specials.flag import Flag, Deleted

__all__ = ['AppendMessage', 'BaseMessage', 'MessageT', 'Policy', 'Policy7Bit']

#: Type variable with an upper bound of :class:`BaseMessage`.
MessageT = TypeVar('MessageT', bound='BaseMessage')

#: :class:`~email.policy.Policy` used for 8-bit serialization.
Policy = SMTP.clone(cte_type='8bit', utf8=True)

#: :class:`~email.policy.Policy` used for 7-bit serialization.
Policy7Bit = Policy.clone(cte_type='7bit')


class AppendMessage(NamedTuple):
    """A single message from the APPEND command.

    Args:
        message: The raw message bytes.
        flag_set: The flags to assign to the message.
        when: The internal timestamp to assign to the message.
        options: The extension options in use for the message.

    """

    message: bytes
    flag_set: FrozenSet[Flag]
    when: datetime
    options: ExtensionOptions


class _BodyOnlyBytesGenerator(BytesGenerator):
    # This should produce a bytestring of a email.message.Message object
    # without including any headers, by exploiting the internal _write_headers
    # method.

    def _write_headers(self, *args, **kwargs):
        pass


class BaseMessage(MessageInterface):
    """Message metadata such as UID, permanent flags, and when the message
    was added to the system.

    Args:
        uid: The UID of the message.
        permanent_flags: Permanent flags for the message.
        internal_date: The internal date of the message.
        expunged: True if this message has been expunged from the mailbox.
        contents: The contents of the message.

    """

    def __init__(self, uid: int, permanent_flags: Iterable[Flag] = None,
                 internal_date: datetime = None, expunged: bool = False,
                 contents: EmailMessage = None, **kwargs: Any) -> None:
        super().__init__()
        self._uid = uid
        self._permanent_flags = set(permanent_flags or [])
        self._internal_date = internal_date
        self._expunged = expunged
        self._contents = contents
        self._kwargs = kwargs

    @classmethod
    def parse(cls: Type[MessageT], uid: int, data: BinaryIO,
              permanent_flags: Iterable[Flag] = None,
              internal_date: datetime = None, expunged: bool = False,
              **kwargs: Any) -> MessageT:
        """Parse the given file object containing a MIME-encoded email message
        into a :class:`BaseLoadedMessage` object.

        Args:
            uid: The UID of the message.
            data: The raw contents of the message.
            permanent_flags: Permanent flags for the message.
            internal_date: The internal date of the message.
            expunged: True if this message has been expunged from the mailbox.

        """
        contents = BytesParser(policy=Policy).parse(data)
        if not isinstance(contents, EmailMessage):
            raise TypeError(contents)
        return cls(uid, permanent_flags, internal_date, expunged,
                   contents, **kwargs)

    @property
    def uid(self) -> int:
        return self._uid

    @property
    def expunged(self) -> bool:
        return self._expunged

    @property
    def permanent_flags(self) -> Set[Flag]:
        return self._permanent_flags

    @property
    def internal_date(self) -> Optional[datetime]:
        return self._internal_date

    @property
    def contents(self) -> EmailMessage:
        """The MIME-parsed message object."""
        if self._contents is None:
            raise ValueError('Message contents not available.')
        return self._contents

    def copy(self: MessageT, new_uid: int) -> MessageT:
        cls = type(self)
        return cls(new_uid, self.permanent_flags, self.internal_date,
                   self.expunged, self._contents, **self._kwargs)

    def get_flags(self, session_flags: SessionFlags = None) -> FrozenSet[Flag]:
        if self.expunged:
            return frozenset({Deleted})
        elif session_flags:
            return frozenset(self.permanent_flags | session_flags[self.uid])
        else:
            return frozenset(self.permanent_flags)

    def update_flags(self, flag_set: Iterable[Flag],
                     flag_op: FlagOp = FlagOp.REPLACE) -> FrozenSet[Flag]:
        new_flags = flag_op.apply(self.permanent_flags, flag_set)
        self.permanent_flags.clear()
        self.permanent_flags.update(new_flags)
        return new_flags

    @classmethod
    def _get_subpart(cls: Type[MessageT], msg: MessageT, section) \
            -> 'EmailMessage':
        if section:
            subpart = msg.contents
            for i in section:
                if subpart.is_multipart():
                    subpart = subpart.get_payload(i - 1)  # type: ignore
                elif i == 1:
                    pass
                else:
                    raise IndexError(i)
            return subpart
        else:
            return msg.contents

    def get_header(self, name: str) -> Sequence[Header]:
        return self.contents.get_all(name, [])

    def get_headers(self, section: Iterable[int] = None,
                    subset: Iterable[str] = None,
                    inverse: bool = False) \
            -> Optional[bytes]:
        try:
            msg = self._get_subpart(self, section)
        except IndexError:
            return None
        ret: List[bytes] = []
        for key, value in msg.items():
            if subset is not None:
                if inverse != (key in subset):
                    ret.append(Policy.fold_binary(key, str(value)))
            else:
                ret.append(Policy.fold_binary(key, str(value)))
        if len(ret) > 0:
            linesep = Policy.linesep.encode('ascii')
            return b''.join(ret) + linesep
        else:
            return None

    def get_body(self, section: Iterable[int] = None,
                 binary: bool = False) -> Optional[bytes]:
        try:
            msg = self._get_subpart(self, section)
        except IndexError:
            return None
        return self._get_bytes(msg, binary)

    def get_text(self, section: Iterable[int] = None,
                 binary: bool = False) -> Optional[bytes]:
        try:
            msg = self._get_subpart(self, section)
        except IndexError:
            return None
        return self._get_bytes(msg, binary, True)

    @classmethod
    def _get_bytes(cls, msg: 'EmailMessage', binary: bool = False,
                   body_only: bool = False) -> bytes:
        ofp = BytesIO()
        generator: Type[BytesGenerator] = _BodyOnlyBytesGenerator \
            if body_only else BytesGenerator
        policy = Policy if binary else Policy7Bit
        try:
            generator(ofp, policy=policy).flatten(msg)  # type: ignore
        except (LookupError, UnicodeEncodeError):
            # Message had a malformed or incorrect charset, default to binary.
            return cls._get_bytes(msg, True, body_only)
        return ofp.getvalue()

    @classmethod
    def _get_size(cls, msg: 'EmailMessage', binary: bool = False) -> int:
        data = cls._get_bytes(msg, binary)
        size = len(data)
        return size

    @classmethod
    def _get_size_with_lines(cls, msg: 'EmailMessage') -> Tuple[int, int]:
        data = cls._get_bytes(msg, True)
        size = len(data)
        lines = data.count(b'\n')
        return size, lines

    def get_size(self, section: Iterable[int] = None,
                 binary: bool = False) -> int:
        try:
            msg = self._get_subpart(self, section)
        except IndexError:
            return 0
        return self._get_size(msg, binary)

    def get_envelope_structure(self) -> EnvelopeStructure:
        return self._get_envelope_structure(self.contents)

    def get_body_structure(self) -> BodyStructure:
        return self._get_body_structure(self.contents)

    @classmethod
    def _get_envelope_structure(cls, msg: EmailMessage) -> EnvelopeStructure:
        return EnvelopeStructure(  # type: ignore
            msg.get('Date'),
            msg.get('Subject'),
            msg.get_all('From'),
            msg.get_all('Sender'),
            msg.get_all('Reply-To'),
            msg.get_all('To'),
            msg.get_all('Cc'),
            msg.get_all('Bcc'),
            msg.get('In-Reply-To'),
            msg.get('Message-Id'))

    @classmethod
    def _get_params(cls, msg: EmailMessage) -> Dict[str, str]:
        content_type = msg.get('Content-Type')
        if content_type:
            return content_type.params  # type: ignore
        else:
            return {}

    @classmethod
    def _get_body_structure(cls, msg: EmailMessage) -> BodyStructure:
        maintype = msg.get_content_maintype()
        subtype = msg.get_content_subtype()
        params = cls._get_params(msg)
        disposition = msg.get('Content-Disposition')
        language = msg.get('Content-Language')
        location = msg.get('Content-Location')
        if msg.is_multipart():
            sub_body_structs = [cls._get_body_structure(part)  # type: ignore
                                for part in msg.get_payload()]  # type: ignore
            return MultipartBodyStructure(  # type: ignore
                subtype, params, disposition, language, location,
                sub_body_structs)
        content_id = msg.get('Content-Id')
        content_desc = msg.get('Content-Description')
        content_encoding = msg.get('Content-Transfer-Encoding')
        if maintype == 'message' and subtype == 'rfc822':
            sub_msg = msg.get_payload(0)
            if not isinstance(sub_msg, EmailMessage):
                raise TypeError(sub_msg)
            sub_env_struct = cls._get_envelope_structure(sub_msg)
            sub_body_struct = cls._get_body_structure(sub_msg)
            size, lines = cls._get_size_with_lines(msg)
            return MessageBodyStructure(  # type: ignore
                params, disposition, language, location, content_id,
                content_desc, content_encoding, None, size, lines,
                sub_env_struct, sub_body_struct)
        elif maintype == 'text':
            size, lines = cls._get_size_with_lines(msg)
            return TextBodyStructure(  # type: ignore
                subtype, params, disposition, language, location,
                content_id, content_desc, content_encoding, None, size, lines)
        size = cls._get_size(msg, True)
        return ContentBodyStructure(  # type: ignore
            maintype, subtype, params, disposition, language, location,
            content_id, content_desc, content_encoding, None, size)
